"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""
from argparse import _ActionsContainer
import tornado
from tornado import gen
import collections
import uuid
import logging
from cosmos.rbac.object import *


_active_roles = None

class RoleCache():
    def __init__(self, *args, **kwargs):
        global _active_roles
        if not _active_roles:
            init_roles()

    def get_roles(self):
        global _active_roles
        return _active_roles

class RbacService():
    ALLOW_ALL_PROPERTY_NAME = "*"
    ALLOW_ALL_OBJECT_NAME = "*"

    def __init__(self, *args, **kwargs):
        self.role_cache = kwargs.get("role_cache", RoleCache())


    def get_roles(self, user):
        active_roles = self.role_cache.get_roles()
        roles = []

        if not user:
            user_role_sids = [ANONYMOUS_USER_ROLE_SID]
        else:
            user_role_sids = user.get("roles", [ANONYMOUS_USER_ROLE_SID])

        for role in active_roles:
            if role.sid in user_role_sids:
                roles.append(role)

        return roles

    def has_owner_access(self, role, object_name, properties, access):
        assert isinstance(role, Role)
        assert isinstance(properties, collections.Iterable)

        required_properties = properties[:]

        for role_item in role.role_items:
            assert isinstance(role_item, RoleItem)

            if (role_item.object_name == RbacService.ALLOW_ALL_OBJECT_NAME or role_item.object_name == object_name) \
                    and access in role_item.owner_access:

                #Check for blanket access
                if RbacService.ALLOW_ALL_PROPERTY_NAME == role_item.property_name:
                    return True

                #We require blanket approval (*) when properties are not specified
                if len(properties) == 0:
                    continue

                if role_item.property_name in required_properties:
                    required_properties.remove(role_item.property_name)

        return len(properties) > 0 and len(required_properties) == 0

    def has_access(self, role, object_name, properties, access):
        assert isinstance(role, Role)
        assert isinstance(properties, collections.Iterable)

        required_properties = properties[:]

        for role_item in role.role_items:
            assert isinstance(role_item, RoleItem)

            if (role_item.object_name == RbacService.ALLOW_ALL_OBJECT_NAME or role_item.object_name == object_name) \
                    and access in role_item.access:

                #Check for blanket access
                if RbacService.ALLOW_ALL_PROPERTY_NAME == role_item.property_name:
                    return True

                #We require blanket approval (*) when properties are not specified
                if len(properties) == 0:
                    continue

                if role_item.property_name in required_properties:
                    required_properties.remove(role_item.property_name)

        return len(properties) > 0 and len(required_properties) == 0


def init_roles():
    global _active_roles
    if not _active_roles:
        _active_roles = []
        for role in WELL_KNOWN_ROLES:
            _active_roles.append(role)

def check_role_item(role_item_defn):
    assert isinstance(role_item_defn, dict)
    property_name = role_item_defn.get("property_name")

    access_list = role_item_defn.get("access", [])
    owner_access_list = role_item_defn.get("owner_access", [])

    if len(access_list) < 1 and len(owner_access_list) < 1:
        raise ValueError("Either access or owner_access items are required")

    for access in access_list:
        if access == AccessType.DELETE and property_name != RbacService.ALLOW_ALL_PROPERTY_NAME:
            raise ValueError("When access type is DELETE property_name must be {0}".format(RbacService.ALLOW_ALL_PROPERTY_NAME))

    for access in owner_access_list:
        if access == AccessType.DELETE and property_name != RbacService.ALLOW_ALL_PROPERTY_NAME:
            raise ValueError("When owner_access type is DELETE property_name must be {0}".format(RbacService.ALLOW_ALL_PROPERTY_NAME))

def get_role_object(role_defn):
    assert isinstance(role_defn, dict)
    assert role_defn.get("type") == "object.Role"

    role_name = role_defn.get("name", None)

    if role_name == None:
        raise ValueError("Role must have name value.")

    role_sid = role_defn.get("sid", None)

    if role_sid == None:
        raise ValueError("Role must have sid value.")

    role_item_defns = role_defn.get("role_items", [])

    role = Role(name=role_name, sid=role_sid)

    for role_item_defn in role_item_defns:
        assert isinstance(role_item_defn, dict)
        check_role_item(role_item_defn)
        role_item = RoleItem(**role_item_defn)
        role.role_items.append(role_item)

    return role


def replace_role(new_role):
    [new_role if existing_role.sid==new_role.sid else existing_role for existing_role in _active_roles]

def clear_non_system_roles():
    global _active_roles
    _active_roles = None
    init_roles()

def update_role_cache(role_defn):
    global _active_roles
    if not _active_roles:
        init_roles()

    try:
        new_role = get_role_object(role_defn)
    except Exception as ex:
        logging.exception(ex)
        return


    for existing_role in _active_roles:
        if existing_role.sid==new_role.sid:
            replace_role(new_role)
            return

    _active_roles.append(new_role)



sample_role = {
    "name": "Administrator",
    "role_items": [
        {
            "access": [
                "INSERT",
                "READ",
                "WRITE"
            ],
            "object_name": "testservice",
            "property_name": "name",
            "type": "object.RoleItem"
        },
        {
            "access": [
                "INSERT",
                "READ",
                "WRITE"
            ],
            "object_name": "testservice",
            "property_name": "address",
            "type": "object.RoleItem"
        },
        {
            "access": [
                "INSERT",
                "READ",
                "WRITE"
            ],
            "object_name": "testservice",
            "property_name": "*",
            "type": "object.RoleItem"
        },
        {
            "access": [
                "DELETE"
            ],
            "object_name": "testservice",
            "property_name": "*",
            "type": "object.RoleItem"
        }
    ],
    "type": "object.Role"
};



if __name__ == "__main__":
    serv = RbacService()

    for role in _active_roles:
        print role.to_JSON()

    role= get_role_object(sample_role)
    role_json =  role.to_JSON()
    role_def = json.loads(role_json)
    role2 = serv.get_role_object(sample_role)
    print role2.to_JSON()

    print serv.has_access(role, "testservice", ["name"], AccessType.READ)
    print serv.has_access(role, "testservice", ["name"], AccessType.UPDATE)
    print serv.has_access(role, "testservice", ["address"], AccessType.READ)




"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from bson import ObjectId
from cosmos.dataservice.objectservice import ObjectService
from cosmos.rbac.object import AccessType, COSMOS_USERS_OBJECT_NAME, SYSTEM_USER, COSMOS_USERS_IDENTITY_OBJECT_NAME
import urllib

__author__ = 'Maruf Maniruzzaman'

import tornado.web
from tornado import gen
from cosmos.service.requesthandler import *
import json
from cosmos.service.utils import MongoObjectJSONEncoder
from constants import *
import tornado.auth
from tornado.concurrent import TracebackFuture, chain_future, return_future

try:
    import urlparse  # py2
except ImportError:
    import urllib.parse as urlparse  # py3

try:
    import urllib.parse as urllib_parse  # py3
except ImportError:
    import urllib as urllib_parse  # py2


hmac_key = None

class LogoutHandler(RequestHandler):
    @gen.coroutine
    def get(self):
        self.logout_current_user()
        self.redirect('/')


def get_jwt_payload_json(id_token):
    jwt_parts = str.split(str(id_token), '.')
    pad = 4 - len(jwt_parts[1]) % 4
    pads=["", "=", "==", "===", ""]
    id_values = base64.b64decode(jwt_parts[1]+pads[pad])

    return json.loads(id_values)


class CosmosAuthHandler(RequestHandler):
    @gen.coroutine
    def _authenticate_user(self, user_identity):
        self.current_identity = user_identity

        identity_type = user_identity.get("identity_type", None)

        #TODO: define methods and override in subclasses so we do not require these ifs and in functions below
        if identity_type == IDENTITY_TYPE_FB_GRAPH:
            fb_userid = user_identity["id"]
            query={"id": fb_userid}
            columns = ["identity_type", "id", "user_id"]
        elif identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = user_identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            sub = id_values.get("sub")
            query = {"sub": sub, "identity_type" : IDENTITY_TYPE_GOOGLE_OAUTH2}
            columns = ["identity_type", "sub", "user_id"]
        else:
            claimed_id = user_identity.get("claimed_id")
            query = {"claimed_id":claimed_id}
            columns = ["identity_type", "claimed_id", "user_id"]

        db = self.settings['db']
        object_service = ObjectService()

        cursor = object_service.find(SYSTEM_USER, db, COSMOS_USERS_IDENTITY_OBJECT_NAME, query, columns)

        if (yield cursor.fetch_next):
            stored_user_id = cursor.next_object()
            user = yield object_service.load(SYSTEM_USER, db, COSMOS_USERS_OBJECT_NAME, str(stored_user_id["user_id"]), [])

            if not user:
                raise tornado.web.HTTPError(500, "User not found for previously stored identity of this user")

            self.set_current_user(user)
        else:
            create_result = yield self._create_user(user_identity)

            cursor = object_service.find(SYSTEM_USER, db, COSMOS_USERS_OBJECT_NAME, {"_id": create_result}, [])
            if (yield cursor.fetch_next):
                user = cursor.next_object()
            else:
                raise tornado.web.HTTPError(500, "Could not find created user")

            result = yield self._create_user_identity(str(user["_id"]), user_identity)

            self.set_current_user(user)


    def _create_user(self, user_identity):
        db = self.settings['db']
        identity_type = user_identity.get("identity_type", None)

        if identity_type == IDENTITY_TYPE_FB_GRAPH:
            fb_userid = user_identity["id"]
            data = {"fb_userid": fb_userid, "roles": []}
        elif identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = user_identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            email = id_values.get("email")
            data = {"email": email, "roles": []}
        else:
            email = user_identity.get("email")
            data = {"email": email, "roles": []}

        object_service = ObjectService()
        return object_service.save(SYSTEM_USER, db, COSMOS_USERS_OBJECT_NAME, data)


    def _create_user_identity(self, user_id, identity):
        db = self.settings['db']
        data = identity
        data["user_id"] = user_id

        identity_type = identity.get("identity_type", None)

        if identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            data.update(id_values)

        object_service = ObjectService()
        return object_service.save(SYSTEM_USER, db, COSMOS_USERS_IDENTITY_OBJECT_NAME, data)

class OpenidLoginHandler(CosmosAuthHandler, tornado.auth.OpenIdMixin):
    @return_future
    def authorize_redirect(self, oauth_scope, callback_uri=None,
                           ax_attrs=["name", "email", "language", "username"],
                           callback=None):

        callback_uri = callback_uri or self.request.uri
        args = self._openid_args(callback_uri, ax_attrs=ax_attrs,
                                 oauth_scope=oauth_scope)
        self.redirect(self._OPENID_ENDPOINT + "?" + urllib_parse.urlencode(args))
        callback()

    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        #TODO: Evaluate security risk of accepting openid.op_endpoint value of remote site response
        self._OPENID_ENDPOINT = self.get_argument("openid.op_endpoint", None)
        if not self._OPENID_ENDPOINT:
            raise tornado.web.HTTPError(400, "Endpoint not found")

        if self.get_argument("openid.mode", None):
            user = yield self.get_authenticated_user()
            if not user:
                raise tornado.web.HTTPError(500, "Openid auth failed")

            assert isinstance(user, dict)

            user['identity_type'] = IDENTITY_TYPE_OPEN_ID

            yield self._authenticate_user(user)
            self.redirect(self.get_argument("next", '/'))
        else:
            callback_uri = self.request.uri
            url = urlparse.urljoin(self.request.full_url(), callback_uri)
            urls = url  # url.replace('http','https')
            yield self.authenticate_redirect(callback_uri=urls)


class GoogleOAuth2LoginHandler(CosmosAuthHandler, tornado.auth.GoogleOAuth2Mixin):
    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):

        #callback_uri = self.request.uri
        #url = urlparse.urljoin(self.request.full_url(), callback_uri)
        #urls = url.replace('http','https')
        urls = self.settings[self._OAUTH_SETTINGS_KEY]['redirect_uri']

        if self.get_argument('code', False):
            # The url must match what was used during authorize phase.
            # It can not accept if there is any mismatch including parameters
            user = yield self.get_authenticated_user(redirect_uri=urls, code=self.get_argument('code'))

            if not user:
                raise tornado.web.HTTPError(500, "Google auth failed")

            user["identity_type"] = IDENTITY_TYPE_GOOGLE_OAUTH2
            yield self._authenticate_user(user)
            self.redirect(self.get_argument("next", '/'))
        else:
            yield self.authorize_redirect(
                redirect_uri=urls,
                client_id=self.settings[self._OAUTH_SETTINGS_KEY]['key'],
                client_secret=self.settings[self._OAUTH_SETTINGS_KEY]['secret'],
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})


def add_params(url, params):
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urllib.urlencode(query)
    return urlparse.urlunparse(url_parts)

class FacebookGraphLoginHandler(CosmosAuthHandler, tornado.auth.FacebookGraphMixin):
    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        nextpg = self.get_argument("next", '/')
        params = {'next':nextpg}
        callback_uri = self.request.uri
        url = urlparse.urljoin(self.request.full_url(), callback_uri)

        redirect_uri = add_params(url, params)

        redirect_uris = redirect_uri  # redirect_uri.replace('http','https')

        if self.get_argument("code", False):
            user = yield self.get_authenticated_user(
                                        redirect_uri= redirect_uris,
                                        client_id=self.settings["facebook_api_key"],
                                        client_secret=self.settings["facebook_secret"],
                                        code=self.get_argument("code")
                                        )
            if user:
                user['identity_type'] = IDENTITY_TYPE_FB_GRAPH
                yield self._authenticate_user(user)
                self.redirect(self.get_argument("next", '/'))
            else:
                raise tornado.web.HTTPError(500, "Facebook auth failed")


        else:
            yield self.authorize_redirect(redirect_uri=redirect_uri,
                              client_id= self.settings["facebook_api_key"],
                              extra_params={"scope": self.settings["facebook_scope"]})

class LoginHandler(RequestHandler):
    @tornado.web.asynchronous
    @gen.coroutine
    def post(self):
        username = self.get_argument("username")
        password = self.get_argument("password")

        if not username or len(username) < 1:
            data = json.loads(self.request.body)
            assert isinstance(data, dict)
            username = data.get("username")
            password = data.get("password")

        db = self.settings['db']
        object_name = COSMOS_USERS_OBJECT_NAME
        obj_serv = ObjectService()

        columns = ["username", "password", "roles"]
        query = {"username": username}

        cursor = obj_serv.find(SYSTEM_USER, db, object_name, query, columns)

        found = yield cursor.fetch_next
        if not found:
            raise tornado.web.HTTPError(401, "Unauthorized")
        else:
            user = cursor.next_object()
            loaded_password_hash = user.get(PASSWORD_COLUMN_NAME)
            hmac_key= self.settings["hmac_key"]
            validate_password(password, loaded_password_hash, hmac_key)
            del user["password"]
            self.set_current_user(user)
            self.redirect('/')


def validate_password(password, saved_password_hash, hmac_key):
    hmac_password = get_hmac_password(password, hmac_key)
    try:
        # FIXME: WARNING possible attack: use hmac.compare_digest(a, b) instead (https://docs.python.org/2/library/hmac.html)
        if not saved_password_hash == hmac_password:
            raise tornado.web.HTTPError(401, "Unauthorized")
    except:
        raise tornado.web.HTTPError(401, "Unauthorized")

def get_hmac_password(password, hmac_key):
    hmac_hex = hmac.new(hmac_key, password).hexdigest()
    hmac_password = "{0}{1}".format(PASSWORD_HMAC_SIGNATURE, hmac_hex)
    return hmac_password

@gen.coroutine
def before_user_insert(db, object_name, data, access_type):
    assert isinstance(hmac_key, object)
    assert object_name == COSMOS_USERS_OBJECT_NAME
    assert isinstance(data, dict)
    assert access_type == AccessType.INSERT or access_type == AccessType.UPDATE

    username = data.get(USERNAME_COLUMN_NAME)

    if username and len(username) > 0:
        username = username.lower()

        data["username"] = username
        object_service = ObjectService()
        query = {"username": username}
        columns=["username"]
        cursor = object_service.find(SYSTEM_USER, db, COSMOS_USERS_OBJECT_NAME, query, columns)

        if(yield cursor.fetch_next):
            user = cursor.next_object()
            if user:
                raise tornado.web.HTTPError(409, "Conflict: Duplicate username")

    password = data.get(PASSWORD_COLUMN_NAME)

    if not password:
        return

    if password.find(PASSWORD_HMAC_SIGNATURE) > 0:
        return

    hmac_password = get_hmac_password(password, hmac_key)

    data[PASSWORD_COLUMN_NAME] = hmac_password



"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from cosmos.service import requesthandler

import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from tornado import gen
import motor
import hashlib

from cosmos.service.utils import MongoObjectJSONEncoder
from cosmos.dataservice.objectservice import *


class GridFSServiceHandler(requesthandler.RequestHandler):

    @tornado.web.asynchronous
    @gen.coroutine
    def get(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        if len(params) < 1 or len(params)>2:
            raise tornado.web.HTTPError(404, "Not found")

        object_name = params[0]

        if not object_name:
            raise tornado.web.HTTPError(404, "Not found")

        object_full_name = object_name

        id = None
        if len(params)==2:
            id = params[1]

        db = self.settings['db']

        preprocessor_list = get_operation_preprocessor(object_full_name, AccessType.READ)
        for preprocessor in preprocessor_list:
            yield  preprocessor(db, object_full_name, id, AccessType.READ)

        result = None
        if id and len(id)>0:
            obj_serv = ObjectService()
            promise = obj_serv.load_file(self.current_user, db, object_full_name, id)
            result = yield promise

            self.write(result.get("body"))
            content_type = result.get("content_type")
            if content_type:
                self.set_header("Content-Type", content_type)
        else:
            list_allowed = self.settings.get("directory_listing_allowed")
            if list_allowed:
                obj_serv = ObjectService()
                promise = obj_serv.list_file(self.current_user, db, object_full_name)
                result = yield promise
                data = {"_d": MongoObjectJSONEncoder().encode(result), "_cosmos_service_array_result_": True};

                self.write(json.dumps(data))
            else:
                raise tornado.web.HTTPError(403, "Directory listing forbidden.")
                """
                self.write('''
                <form enctype="multipart/form-data" method="POST">
                    File tu upload: <input name="uploadedfile" type="file" /><br />
                    <input type="submit" value="Upload File" />
                </form>''')
                """

        post_processor_list = get_operation_postprocessor(object_full_name, AccessType.READ)
        for post_processor in post_processor_list:
            yield post_processor(db, object_full_name, result, AccessType.READ)

        self.finish()

    @tornado.web.asynchronous
    @gen.coroutine
    def post(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        object_name = params[0]

        object_full_name = object_name
        files = self.request.files["uploadedfile"]

        db = self.settings['db']
        obj_serv = ObjectService()
        result_list = []
        for file in files:
            preprocessor_list = get_operation_preprocessor(object_full_name, AccessType.INSERT)
            for preprocessor in preprocessor_list:
                yield preprocessor(db, object_full_name, file, AccessType.INSERT)

            promise = obj_serv.save_file(self.current_user, db, object_full_name, file)
            result = yield promise

            result_list.append(result)

            post_processor_list = get_operation_postprocessor(object_full_name, AccessType.INSERT)
            for post_processor in post_processor_list:
                yield  post_processor(db, object_full_name, result, AccessType.INSERT)

        data = {"_d": MongoObjectJSONEncoder().encode(result_list), "_cosmos_service_array_result_": True};

        self.write(data)
        self.finish()

    @tornado.web.asynchronous
    @gen.coroutine
    def delete(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        object_name = params[0]

        if not object_name or len(object_name) < 1:
            raise tornado.web.HTTPError(404, "Not found")

        object_full_name = object_name

        file_id = None
        if len(params) == 2:
            file_id = params[1]
        else:
            raise tornado.web.HTTPError(400, "File id required")


        if not file_id or len(file_id)<1:
            raise tornado.web.HTTPError(400, "File id required")

        db = self.settings['db']

        preprocessor_list = get_operation_preprocessor(object_full_name, AccessType.DELETE)
        for preprocessor in preprocessor_list:
            yield preprocessor(object_full_name, None, AccessType.DELETE)

        obj_serv = ObjectService()
        yield obj_serv.delete_file(self.current_user, db, object_full_name, file_id)


        post_processor_list = get_operation_postprocessor(object_full_name, AccessType.DELETE)
        for post_processor in post_processor_list:
            yield  post_processor(db, object_full_name, None, AccessType.DELETE)

        self.write('{"result":"OK"}')
        self.finish()



"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import os
import sys
import logging
import getpass
from tornado import gen
import createproject

from cosmos.dataservice.objectservice import *

class CommandHandler():
    def __init__(self, *args, **kwargs):
        self.db = kwargs.get("db", None)

    def handle_command(self, *args, **kwarg):
        current_directory = args[0]
        command = args[1]
        command_args = args[2]

        print command

        if command == "new-admin":
            self.create_admin_user()
            return
        if command == "add-herokusettings":
            add_heroku_settings(current_directory)
            sys.exit(0)
        elif command == "new-project":
            project_type = command_args.get("arg0", None)
            createproject.new_project(current_directory, project_type)
            sys.exit(0)
        else:
            print_usage()
            sys.exit(0)

    def get_input(self, prompt):
        input = None
        while not input or len(input)==0:
            input = raw_input(prompt).strip()

        return input

    def create_user(self, username, password, email, roles):
        data = {"username": username, "password": password, "email": email, "roles": roles}
        object_service = ObjectService()
        result = object_service.save(SYSTEM_USER, self.db, COSMOS_USERS_OBJECT_NAME, data)
        logging.info("Admin user was created successfully.")
        return result

    def create_admin_user(self):
        username = self.get_input('Enter admin username: ')
        password = getpass.getpass('Enter admin password: ')
        password_re = getpass.getpass('Repeat admin password: ')
        if password != password_re:
            print "Password mismatch"
            sys.exit(1)

        email = self.get_input('Enter admin email: ')
        self.method_name(username, password, email, [ADMIN_USER_ROLE_SID])
        sys.exit(0)

def print_usage():
    print "Unknown command.\ncosmosadmin new-admin\ncosmosadmin new-project [angular]\ncosmosadmin add-heroku-settings\n"

def add_heroku_settings(current_directory):
    proc_file_path = os.path.join(current_directory, "Procfile")
    with open(proc_file_path, 'w') as procfile:
            procfile.write("web: python cosmosmain.py start-service $PORT")
    req_file_path = os.path.join(current_directory, "requirements.txt")

    with open(req_file_path, 'w') as req_file:
            req_file.write("cosmos")


def admin_main():
    current_directory = os.getcwd()
    if len(sys.argv) < 1:
        print_usage()
        return
    else:
        command = sys.argv[1].strip()
        arg0=None
        arg1=None
        arg2=None
        try:
            arg0 = sys.argv[2]
            arg1 = sys.argv[3]
            arg2 = sys.argv[4]
        except:
            pass

        handler = CommandHandler()
        handler.handle_command(current_directory, command, {"arg0":arg0, "arg1":arg1, "arg2":arg2})

    tornado.ioloop.IOLoop.instance().start()



"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

PASSWORD_HMAC_SIGNATURE = "hmac:"
PASSWORD_COLUMN_NAME = "password"
USERNAME_COLUMN_NAME = "username"
USER_COOKIE_NAME = "usersecret"
USER_PLAINTEXT_COOKIE_NAME = "user"
IDENTITY_TYPE_FB_GRAPH = 'fb_graph'
IDENTITY_TYPE_GOOGLE_OAUTH2 = "google_oauth2"
IDENTITY_TYPE_OPEN_ID = "openid"


"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from cosmos.service import requesthandler

import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from tornado import gen

from cosmos.service.utils import MongoObjectJSONEncoder
from cosmos.dataservice.objectservice import *



class ServiceHandler(requesthandler.RequestHandler):

    @tornado.web.asynchronous
    @gen.coroutine
    def get(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        if len(params) < 1 or len(params)>2:
            raise tornado.web.HTTPError(404, "Not found")

        object_name = params[0]
        if not object_name:
            raise tornado.web.HTTPError(404, "Not found")

        id = None
        if len(params)==2:
            id = params[1]

        db = self.settings['db']

        obj_serv = ObjectService()
        columns_str = self.get_argument("columns", None)
        filter_str = self.get_argument("filter", None)

        if filter_str:
            query=json.loads(filter_str)
        else:
            query=None

        if columns_str:
            columns = columns_str.split(',')
            columns = [column.strip() for column in columns]
        else:
            columns = []

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.READ)
        for preprocessor in preprocessor_list:
            yield  preprocessor(db, object_name, query, AccessType.READ)

        result = None
        if id and len(id)>0:
            cursor = obj_serv.load(self.current_user, db, object_name, id, columns)
            result = yield cursor
            if not result:
                raise tornado.web.HTTPError(404, "Not found")
            data = MongoObjectJSONEncoder().encode(result)
        else:
            cursor = obj_serv.find(self.current_user, db, object_name, query, columns)
            result_list = []
            while(yield cursor.fetch_next):
                qry_result=cursor.next_object()
                result_list.append(qry_result)
            result = result_list
            data = {"_d": MongoObjectJSONEncoder().encode(result_list), "_cosmos_service_array_result_": True};

        post_processor_list = get_operation_postprocessor(object_name, AccessType.READ)
        for post_processor in post_processor_list:
            yield post_processor(db, object_name, result, AccessType.READ)

        self.content_type = 'application/json'
        self.write(data)
        self.finish()

    @tornado.web.asynchronous
    @gen.coroutine
    def post(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        object_name = params[0]
        try:
            data = json.loads(self.request.body)
            assert isinstance(data, dict)
        except ValueError, ve:
            raise tornado.web.HTTPError(400, ve.message)

        db = self.settings['db']

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.INSERT)
        for preprocessor in preprocessor_list:
            yield preprocessor(db, object_name, data, AccessType.INSERT)

        obj_serv = ObjectService()
        promise = obj_serv.save(self.current_user, db, object_name,data)
        result = yield promise
        data = MongoObjectJSONEncoder().encode(result)

        post_processor_list = get_operation_postprocessor(object_name, AccessType.INSERT)
        for post_processor in post_processor_list:
            yield  post_processor(db, object_name, result, AccessType.INSERT)

        self.write(data)
        self.finish()

    @gen.coroutine
    @tornado.web.asynchronous
    def put(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        object_name = params[0]
        id = params[1]
        try:
            data = json.loads(self.request.body)
            assert isinstance(data, dict)
        except ValueError, ve:
            raise tornado.web.HTTPError(400, ve.message)

        db = self.settings['db']

        data['modifytime'] = str(datetime.datetime.now())

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.UPDATE)
        for preprocessor in preprocessor_list:
            yield preprocessor(db, object_name, data, AccessType.UPDATE)

        obj_serv = ObjectService()
        promise = obj_serv.update(self.current_user, db, object_name, id, data)
        result = yield promise
        data = MongoObjectJSONEncoder().encode({"error": result.get("err"),  "n":result.get("n"), "ok": result.get("ok"), "updatedExisting": result.get("updatedExisting")})

        post_processor_list = get_operation_postprocessor(object_name, AccessType.UPDATE)
        for post_processor in post_processor_list:
            yield  post_processor(db, object_name, result, AccessType.UPDATE)

        self.write(data)
        self.finish()

    @tornado.web.asynchronous
    @gen.coroutine
    def delete(self, object_path):
        params = object_path.split('/')
        params = filter(len, params)
        object_name = params[0]
        id = params[1]

        db = self.settings['db']

        preprocessor_list = get_operation_preprocessor(object_name, AccessType.DELETE)
        for preprocessor in preprocessor_list:
            yield  preprocessor(object_name, None, AccessType.DELETE)

        obj_serv = ObjectService()
        promise = obj_serv.delete(self.current_user, db, object_name, id)
        result = yield promise
        data = MongoObjectJSONEncoder().encode({"error": result.get("err"),  "n":result.get("n"), "ok": result.get("ok")})

        post_processor_list = get_operation_postprocessor(object_name, AccessType.DELETE)
        for post_processor in post_processor_list:
            yield  post_processor(db, object_name, result, AccessType.DELETE)

        self.write(data)
        self.finish()









"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from logging import exception
from tornado import gen

from tornado.httputil import url_concat
import tornado.websocket
import tornado.web
from pymongo import MongoClient
from pymongo import Connection
import time
import logging
import threading
import json
import bson
import time
import math

OPLOG_COLLECTION = 'oplog.rs'
REPLSET_COLLECTION = 'system.replset'

"""
SETUP

In /etc/mongodb.conf add the following line to enable replication

replSet = rs0

Then from mongo console do rs.initiate()

> rs.initiate()
{
	"info2" : "no configuration explicitly specified -- making one",
	"me" : "mongodb2:27017",
	"info" : "Config now saved locally.  Should come online in about a minute.",
	"ok" : 1
}
Now do er.config() to see the status

> rs.config()
{
	"_id" : "rs0",
	"version" : 1,
	"members" : [
		{
			"_id" : 0,
			"host" : "mongodb2:27017"
		}
	]
}

"""

subscribers = []
continue_monitor = False

class ChangeRequestHandler(tornado.web.RequestHandler):
    @tornado.web.asynchronous
    def get(self):
        ns = self.get_argument("ns")
        logging.debug("ChangeRequestHandler ns={0}".format(ns))
        for sub in subscribers:
            assert isinstance(sub, ChangeMonitor)
            sub.handle_change(ns)

        self.content_type = 'application/json'
        data = '{"result":"OK"}'
        self.write(data)
        self.finish()


class ChangeMonitor(tornado.websocket.WebSocketHandler):
    def open(self):
        logging.debug("WebSocket opened")
        self.ns = []
        subscribers.append(self)

    def on_message(self, message):
        logging.debug("Message received: " + message)
        message = json.loads(message)
        assert isinstance(message, dict)
        if message["type"] == "monitor_ns":
            ns = message["ns"]
            logging.info("New client is subscribing for namespace {0}".format(ns))
            self.ns.append(ns)

    def handle_change(self, change_ns):
        if change_ns in self.ns:
            self.write_message(change_ns)

    def on_close(self):
        logging.debug("Websocket closed")
        subscribers.remove(self)


def report_change(report_server, doc):
    try:
        params = {"ns": doc["ns"]}
        url = url_concat(report_server, params)
        http_client = tornado.httpclient.HTTPClient()
        http_client.fetch(url)
    except Exception, ex:
        logging.exception(ex)


def monitor(params):
    db_uri = params.get("db_uri", "127.0.0.1")
    report_server = params["report_server"]
    end_callback = params.get("end_callback", None)
    logging.info("Monitor started for database at {0}. Will report to {1}".format(db_uri, report_server))
    client = MongoClient(db_uri)

    db = client.local
    coll = db[OPLOG_COLLECTION]

    # @@TODO: Find the highest timestamp instead of current time
    # coll.find({},{ts:1}).sort({$natural:-1}).limit(1).toArray()

    ts_now = int(math.floor(time.time()))
    ts = bson.timestamp.Timestamp(ts_now, 1)
    cursor = coll.find({"ts": {"$gt": ts}}, tailable=True, awaitdata=True)
    while cursor.alive and continue_monitor:
        try:
            doc = cursor.next()
            logging.debug(doc)
            report_change(report_server, doc)
        except StopIteration:
            time.sleep(1)

    if end_callback:
        end_callback()


def start_object_change_monitor(report_server, db_uri, end_callback = None):
    worker = threading.Thread(target=monitor,
                              args=({"db_uri": db_uri, "report_server": report_server, "end_callback": end_callback},))
    worker.start()
    return worker
